# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import time
import inspect
import importlib
import builtins
import functools
import textwrap

from .message import Message, dumps
from .name import basename
from .objects import OK, Fail, Error, Skip, Null
from .objects import XOK, XFail, XError, XNull
from .objects import Value, Metric, Ticket, Attribute, Tag, Requirement, Node
from .parallel import top, current, previous

def current_dir(frame=None):
    """Return directory of the current source file.
    """
    if frame is None:
        frame = inspect.currentframe().f_back
    return os.path.dirname(os.path.abspath(frame.f_globals["__file__"]))

def current_module(frame=None):
    """Return reference to the current module.
    """
    if frame is None:
        frame = inspect.currentframe().f_back
    return sys.modules[frame.f_globals["__name__"]]

def load_module(name, package=None):
    """Load module by name.
    """
    return importlib.import_module(module, package=package)

def load(name, test=None, package=None, frame=None):
    """Load test by name from module.

    :param name: module name or module
    :param test: test class or method name to load (optional)
    :param package: package if module name is relative (optional)
    """
    if name is None or name == ".":
        if frame is None:
            frame = inspect.currentframe().f_back
        module = sys.modules[frame.f_globals["__name__"]]
    elif inspect.ismodule(name):
        module = name
    else:
        module = importlib.import_module(name, package=package)
    if test:
        test = getattr(module, test, None)
    if test is None:
        test = getattr(module, "TestCase", None)
    if test is None:
        test = getattr(module, "TestSuite", None)
    if test is None:
        test = getattr(module, "Test", None)
    return test

def append_path(pathlist, path, *rest, **kwargs):
    """Append path relative to the caller
    to the path list.

    :param pathlist: path list
    :param path: path relative to the caller
    :param *rest: rest of the path
    :param pos: insert at given position,
       default: append to the end of the list
    """
    pos = kwargs.pop("pos", None)
    frame = inspect.currentframe().f_back
    dir = os.path.join(os.path.dirname(os.path.abspath(frame.f_globals["__file__"])), path, *rest)
    if dir not in pathlist:
        if pos is None:
            pathlist.append(dir)
        else:
            pathlist.insert(pos, dir)
    return pathlist

def main(frame=None):
    """Return true if caller is the main module.

    :param frame: caller frame
    """
    if frame is None:
        frame = inspect.currentframe().f_back
    return frame.f_globals["__name__"] == "__main__"

class args(dict):
    pass

def attribute(name, value, type=None, group=None, uid=None, base=Attribute, test=None):
    obj = base(name=name, value=value, type=type, group=group, uid=uid)
    if test is None:
        test = current()
    if test.attributes.get(obj.name, None) is not None:
        raise NameError(f"attribute named '{obj.name}' already exists")
    test.attributes[obj.name] = obj
    test.io.output.attribute(obj)

def requirement(name, version, description=None, link=None,
        priority=None, type=None, group=None, uid=None, base=Requirement, test=None):
    obj = base(name=name, version=version, description=description, link=link,
        priority=priority, type=type, group=group, uid=uid)
    if test is None:
        test = current()
    if test.requirements.get(obj.name, None) is not None:
        raise NameError(f"requirement named '{obj.name}' already exists")
    test.requirements[obj.name] = obj
    test.io.output.requirement(obj)

def tag(value, test=None):
    value = str(value)
    if test is None:
        test = current()
    test.tags.add(value)
    test.io.output.tag(Tag(value))

def metric(name, value, units, type=None, group=None, uid=None, base=Metric, test=None):
    obj = base(name=name, value=value, units=units, type=type, group=group, uid=uid)
    if test is None:
        test = current()
    test.result.metrics.append(obj)
    test.io.output.metric(obj)

def ticket(name, link=None, type=None, group=None, uid=None, base=Ticket, test=None):
    obj = base(name=name, link=link, type=type, group=group, uid=uid)
    if test is None:
        test = current()
    test.result.tickets.append(obj)
    test.io.output.ticket(obj)

def value(name, value, type=None, group=None, uid=None, base=Value, test=None):
    obj = base(name=name, value=value, type=type, group=group, uid=uid)
    if test is None:
        test = current()
    test.result.values.append(obj)
    test.io.output.value(obj)
    return value

def private_key(value=None, test=None):
    if test is None:
        test = current()
    if value is not None:
        test.private_key = value
    else:
        value = test.private_key
    return value

def text(*message, start="", end="\n", format=None, dedent=True, strip=False, file=None, test=None):
    if test is None:
        test = current()

    text = "\n".join(message)

    if file is None:
        file = getattr(current().context, "file", None)

    if dedent:
        text = textwrap.dedent(text)

    if strip:
        text = text.strip()

    if format:
        text = format.format(text=text)

    text = f"{start}{text}{end}"

    if file:
        file.write(text)

    test.io.output.text(text)

def note(message, test=None):
    if test is None:
        test = current()
    test.io.output.note(message)

def debug(message, test=None):
    if test is None:
        test = current()
    test.io.output.debug(message)

def trace(message, test=None):
    if test is None:
        test = current()
    test.io.output.trace(message)

def message(message, test=None, stream=None):
    if test is None:
        test = current()
    test.io.output.message(Message.NONE, {"message": str(message)}, stream=stream)

def exception(exc_type=None, exc_value=None, exc_traceback=None, test=None):
    if test is None:
        test = current()
    test.io.output.exception(exc_type, exc_value, exc_traceback)

def result(type, *args, test=None):
    if type is OK:
        return ok(*args, test=test)
    elif type is Fail:
        return fail(*args, test=test)
    elif type is Skip:
        return skip(*args, test=test)
    elif type is Error:
        return err(*args, test=test)
    elif type is Null:
        return null(*args, test=test)
    elif type is XOK:
        return xok(*args, test=test)
    elif type is XFail:
        return xfail(*args, test=test)
    elif type is XNull:
        return xnull(*args, test=test)
    raise TypeError(f"invalid result type {type}")

def cleanup(func, *args, test=None, context=None, **kwargs):
    if context is None:
        if test is None:
            test = current()
        context = test.context
    context.cleanup(func, *args, **kwargs)

def ok(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = OK(test=test.name, message=message, reason=reason)
    raise test.result

def fail(message=None, reason=None, test=None, type=None):
    if test is None:
        test = current()

    if type is not None:
        if not message:
            raise ValueError("message must be specified")
        with type(message):
           fail(reason=reason)

    test.result = Fail(test=test.name, message=message, reason=reason)
    raise test.result

def skip(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = Skip(test=test.name, message=message, reason=reason)
    raise test.result

def err(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = Error(test=test.name, message=message, reason=reason)
    raise test.result

def null(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = Null(test=test.name, message=message, reason=reason)
    raise test.result

def xok(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = XOK(test=test.name, message=message, reason=reason)
    raise test.result

def xfail(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = XFail(test=test.name, message=message, reason=reason)
    raise test.result

def xerr(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = XError(test=test.name, message=message, reason=reason)
    raise test.result

def xnull(message=None, reason=None, test=None):
    if test is None:
        test = current()
    test.result = XNull(test=test.name, message=message, reason=reason)
    raise test.result

def pause(message=None, test=None):
    if test is None:
        test = current()
    if message is None:
        message = ", e"
    else:
        message = message.strip()
        message = f", {message[:1].lower() + message[1:]}\nE"
    test.io.output.prompt(f"Paused{message}nter any key to continue...")
    builtins.input()
    test.io.output.input("")

def input(type, multiline=False, choices=None, confirm=True, test=None):
    nl = "\n"

    if test is None:
        test = current()

    def confirmed():
        prompt = "Is this correct [Y/n]? "
        test.io.output.prompt(prompt)
        answer = builtins.input()
        test.io.output.input(answer or "Y")
        return answer in ["Y", "y", ""]

    def multilined():
        lines = []
        while True:
            try:
                lines.append(builtins.input())
            except EOFError:
                break
        return '\n'.join(lines)

    if builtins.type(type) is str:
        while True:
            test.io.output.prompt(f"{type.strip()}{(nl + '(Press Enter then Ctrl-D to finish)') if multiline else ''}\n\n"
                f"{('[' + ','.join([repr(c) for c in choices]) + ']' + nl*2) if choices else ''}")
            if multiline:
                text = multilined()
            else:
                text = builtins.input()

            test.io.output.input(text)

            if choices:
                if text not in choices:
                    continue

            if confirm:
                if confirmed():
                    break
            else:
                break

        return text

    elif type is result:
        input_results_map = {
            "ok": ok,
            "": ok,
            "fail": fail,
            "error": err,
            "null": null,
            "skip": skip,
            "xfail": functools.partial(xfail, None),
            "xerror": functools.partial(xerr, None),
            "xok": functools.partial(xok, None),
            "xnull": functools.partial(xnull, None),
        }

        while True:
            test.io.output.prompt(f"Enter `{basename(test.name)}` result? ")
            input_result = builtins.input()
            test.io.output.input(input_result or "OK")

            if input_results_map.get(input_result.split(" ", 1)[0].lower()) is None:
                continue

            if confirm:
                if confirmed():
                    break
            else:
                break

        message_or_reason = None

        if " " in input_result:
            message_or_reason = input_result.split(" ", 1)[-1]

        input_results_map.get(input_result.split(" ", 1)[0].lower())(message_or_reason)

    else:
        raise ValueError(f"invalid type {type}")

def getsattr(obj, name, *default):
    """Get attribute or set it to the default value.
    """
    value = getattr(obj, name, *default)
    setattr(obj, name, value)
    return value

def current_time(test=None):
    """Return current execution time.
    """
    if test is None:
        test = current()

    if test.test_time is None:
       return time.time() - test.start_time
    else:
       return test.test_time




#  Copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""ASN.1 definitions.

Not all ASN.1-handling code use these definitions, but when it does, they should be here.
"""

from testflows._core.contrib.pyasn1.type import univ, namedtype, tag


class PubKeyHeader(univ.Sequence):
    componentType = namedtype.NamedTypes(
            namedtype.NamedType('oid', univ.ObjectIdentifier()),
            namedtype.NamedType('parameters', univ.Null()),
    )


class OpenSSLPubKey(univ.Sequence):
    componentType = namedtype.NamedTypes(
            namedtype.NamedType('header', PubKeyHeader()),

            # This little hack (the implicit tag) allows us to get a Bit String as Octet String
            namedtype.NamedType('key', univ.OctetString().subtype(
                    implicitTag=tag.Tag(tagClass=0, tagFormat=0, tagId=3))),
    )


class AsnPubKey(univ.Sequence):
    """ASN.1 contents of DER encoded public key:

    RSAPublicKey ::= SEQUENCE {
         modulus           INTEGER,  -- n
         publicExponent    INTEGER,  -- e
    """

    componentType = namedtype.NamedTypes(
            namedtype.NamedType('modulus', univ.Integer()),
            namedtype.NamedType('publicExponent', univ.Integer()),
    )



# Copyright 2021 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# to the end flag
import atexit
import weakref
import itertools
import threading
import concurrent.futures.thread as _base

from .future import Future 
from .. import _get_parallel_context
from ..asyncio import is_running_in_event_loop, asyncio

_tasks_queues = weakref.WeakKeyDictionary()
_shutdown = False

def _python_exit():
    global _shutdown
    _shutdown = True
    # FIXME: debug leaked
    #import sys
    #import gc
    #for k in (list(_tasks_queues.keys())):
    #    print(f"{k} {sys.getrefcount(k)} {gc.get_referrers(k)}")
    for work_queue in _tasks_queues.values():
        if work_queue._loop.is_running():
            asyncio.run_coroutine_threadsafe(work_queue.put(None), loop=work_queue._loop)
    for task in _tasks_queues.keys():
        task.result()

atexit.register(_python_exit)

class _AsyncWorkItem(_base._WorkItem):
    async def run(self):
        if not self.future.set_running_or_notify_cancel():
            return
        try:
            result = self.fn(*self.args, **self.kwargs)
            if asyncio.iscoroutine(result):
                result = await result
        except BaseException as exc:
            self.future.set_exception(exc)
            self = None
        else:
            self.future.set_result(result)


async def _worker(executor_weakref, work_queue):
    while True:
        work_item = await work_queue.get()
        try:
            if work_item is not None:
                await work_item.run()
                del work_item
                continue
        finally:
            work_queue.task_done()

        executor = executor_weakref()
        try:
            if _shutdown or executor is None or executor._shutdown:
                if executor is not None:
                    executor._shutdown = True
                await work_queue.put(None)
                return
        finally:
            del executor


async def _loop_send_stop_event(stop_event):
    """Set stop event to stop event loop.
    """
    stop_event.set()


async def _loop_main(stop_event):
    """Event loop main function
    that blocks until stop event is set.
    """
    await stop_event.wait()


def _async_loop_thread(loop, stop_event):
    """Async loop thread.
    """
    asyncio.set_event_loop(loop)

    loop.run_until_complete(_loop_main(stop_event))

    tasks = [task for task in asyncio.all_tasks(loop=loop)]
    for task in tasks:
        task.cancel()

    loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
    loop.close()


class AsyncPoolExecutor(_base._base.Executor):
    """Async pool executor.
    """
    _counter = itertools.count().__next__

    def __init__(self, max_workers=1024, task_name_prefix="",
            loop=None, _check_max_workers=True):
        if _check_max_workers and int(max_workers) <= 0:
            raise ValueError("max_workers must be greater than 0")
        self._open = False
        self._max_workers = max_workers
        self._loop = loop or asyncio.new_event_loop()
        self._loop_stop_event = asyncio.Event(loop=self._loop)
        self._work_queue = asyncio.Queue(loop=self._loop)
        self._tasks = set()
        self._shutdown = False
        self._shutdown_lock = threading.Lock()
        self._task_name_prefix = (task_name_prefix or
            ("AsyncPoolExecutor-%d" % self._counter()))
        self._async_loop_thread = None

    @property
    def open(self):
        return bool(self._open)

    def __enter__(self):
        with self._shutdown_lock:
            if not self._open:
                self._async_loop_thread = threading.Thread(target=_async_loop_thread,
                    kwargs={"loop": self._loop, "stop_event": self._loop_stop_event},
                    daemon=True)
                self._async_loop_thread.start()
                self._open = True
            return self

    def submit(self, fn, args=None, kwargs=None, block=True):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}

        with self._shutdown_lock:
            if not self._open:
                raise RuntimeError("cannot schedule new futures before pool is opened")
            if self._shutdown:
                raise RuntimeError("cannot schedule new futures after shutdown")
            if _shutdown:
                raise RuntimeError("cannot schedule new futures after interpreter shutdown")

            future = Future()
            ctx = _get_parallel_context()
            args = fn, *args
            work_item = _AsyncWorkItem(future, ctx.run, args, kwargs)

            idle_workers = self._adjust_task_count()

            if (idle_workers or block) and self._max_workers > 0:
                if is_running_in_event_loop() and asyncio.get_event_loop() is self._loop:
                    raise RuntimeError("deadlock detected")
                asyncio.run_coroutine_threadsafe(self._work_queue.put(work_item), loop=self._loop).result()

        if (not block and not idle_workers) or self._max_workers < 1:
            if is_running_in_event_loop() and asyncio.get_event_loop() is self._loop:
                raise RuntimeError("deadlock detected")
            asyncio.run_coroutine_threadsafe(work_item.run(), loop=self._loop).result()

        if is_running_in_event_loop():
            return asyncio.wrap_future(future)

        return future

    def _adjust_task_count(self):
        """Increase worker count up to max_workers if needed.
        Returns `True` if worker is immediately available
        to handle the work item or `False` otherwise.
        """
        if len(self._tasks) - self._work_queue._unfinished_tasks > 0:
            return True

        def weakref_cb(_, work_queue=self._work_queue):
            asyncio.run_coroutine_threadsafe(work_queue.put(None), loop=self._loop).result()

        num_tasks = len(self._tasks)
        if num_tasks < self._max_workers:
            task_name = "%s_%d" % (self._task_name_prefix or self, num_tasks)
            task = asyncio.run_coroutine_threadsafe(_worker(weakref.ref(self, weakref_cb), self._work_queue), loop=self._loop)
            self._tasks.add(task)
            _tasks_queues[task] = self._work_queue
            return True

        return False

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.shutdown(wait=True)
        return False

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            if self._shutdown:
                return
            self._shutdown = True
            if not self._loop.is_running():
                return
            asyncio.run_coroutine_threadsafe(
                self._work_queue.put(None), loop=self._loop).result()

        if wait:
            exc = None
            for task in self._tasks:
                try:
                    task.result()
                except BaseException as e:
                    exc = e
            try:
                if exc is not None:
                    raise exc
            finally:
                asyncio.run_coroutine_threadsafe(
                        _loop_send_stop_event(self._loop_stop_event), loop=self._loop
                    ).result()
                if self._async_loop_thread is not None:
                    self._async_loop_thread.join()


class SharedAsyncPoolExecutor(AsyncPoolExecutor):
    """Shared async pool executor.
    """
    def __init__(self, max_workers, task_name_prefix=""):
        self.initargs = (max_workers, task_name_prefix) 

        if int(max_workers) < 0:
            raise ValueError("max_workers must be positive or 0")
        super(SharedAsyncPoolExecutor, self).__init__(
            max_workers=max_workers-1, task_name_prefix=task_name_prefix, _check_max_workers=False)

    def submit(self, fn, args=None, kwargs=None, block=False):
        return super(SharedAsyncPoolExecutor, self).submit(fn=fn, args=args, kwargs=kwargs, block=block)


GlobalAsyncPoolExecutor = SharedAsyncPoolExecutor



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from testflows._core.contrib.arpeggio import RegExMatch as _
from testflows._core.contrib.arpeggio import OneOrMore, ZeroOrMore, EOF, Optional, Not
from testflows._core.contrib.arpeggio import ParserPython as PEGParser
from testflows._core.contrib.arpeggio import PTNodeVisitor, visit_parse_tree

class Visitor(PTNodeVisitor):
    def __init__(self, *args, **kwargs):
        self.header_ids = {}
        self.start_after = kwargs.pop("start_after", None)
        self.started = True
        if self.start_after is not None:
            self.started = False
        self.levels = []
        self.current_level = 0
        self.output = []
        super(Visitor, self).__init__(*args, **kwargs)

    def process_heading(self, node, children):
        level = node[0].value.count("#")
        # only include in TOC levels 2 and higher
        if level < 2:
            return None
        if not self.started:
            if self.start_after in node.heading_name.value:
                self.started = True
                return None
        if not self.started:
            return None
        # normalize header level
        level -= 1
        if self.current_level < level:
            self.levels = self.levels[:level - 1]
        if len(self.levels) < level:
            self.levels += [0] * (level - len(self.levels))
        self.current_level = level
        self.levels[self.current_level - 1] += 1
        num = '.'.join([str(l) for l in self.levels[:self.current_level]])
        return level, num

    def visit_heading(self, node, children):
        res = self.process_heading(node, children)
        if not res:
            return
        level, num = res
        name = node.heading_name.value
        anchor = re.sub(r"\s+", "-", re.sub(r"[^a-zA-Z0-9-_\s]+", "", name.lower()))
        # handle duplicate header ids
        if self.header_ids.get(anchor) is None:
            self.header_ids[anchor] = 1
        else:
            anchor = f"{anchor}{str(self.header_ids[anchor])}"
            self.header_ids[anchor] += 1
        indent = "  " * (level - 1)
        self.output.append(f"{indent}* {'.'.join([str(l) for l in self.levels[:self.current_level]])} [{name}](#{anchor})")

    def visit_document(self, node, children):
        self.output = "\n".join(self.output)
        if self.output:
            self.output += "\n"
        return self.output or None


class UpdateVisitor(PTNodeVisitor):
    def __init__(self, *args, **kwargs):
        self.heading = kwargs.pop("heading", None)
        self.toc = kwargs.pop("toc", "")
        self.output = []
        # states
        self.state_inside_toc = False
        self.state_done = False
        super(UpdateVisitor, self).__init__(*args, **kwargs)

    def visit_line(self, node, children):
        if not self.state_inside_toc:
            self.output.append(str(node))

    def visit_heading(self, node, children):
        self.output.append("".join(children))
        if not self.state_done and node.heading_name.value == self.heading:
            self.state_inside_toc = True
            self.output.append("\n")
            self.output.append(self.toc)
        else:
            if self.state_inside_toc:
                self.state_done = True
            self.state_inside_toc = False

    def visit_document(self, node, children):
        self.output = "".join(self.output)
        return self.output or None


def Parser():
    """Returns markdown heading parser.
    """
    def line():
        return _(r"[^\n]*\n")

    def heading():
        return [
            (_(r"\s*#+\s+"), heading_name, _(r"\n?")),
            (heading_name, _(r"\n?[-=]+\n?"))
        ]

    def heading_name():
        return _(r"[^\n]+")

    def document():
        return Optional(OneOrMore([heading, line])), EOF

    return PEGParser(document, skipws=False)


def generate(source, destination, heading, update=False):
    """Generate TOC for markdown source.

    :param source: source file-like object
    :param destination: destination file-like object
    :param heading: table of contents heading name
    :param update: write to destination source with the updated table of contents, default: False
    """
    parser = Parser()
    source_data = source.read()
    tree = parser.parse(source_data)
    toc = visit_parse_tree(tree, Visitor(start_after=heading))
    destination_data = ""
    if update:
        destination_data = visit_parse_tree(tree, UpdateVisitor(heading=heading, toc=toc))
    else:
        destination_data = toc
    if destination_data:
        destination.write(destination_data)



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def wstrip(s, word, left=True, right=True):
    """Strip word from the beginning or the end of the string.
    By default strips from both sides.
    """
    step = len(word)
    start_pos = None
    end_pos = None

    if not word:
        return s

    while True:
        found = False
        if left and s.startswith(word, start_pos, end_pos):
            if start_pos is None:
                start_pos = 0
            start_pos += step
            found = True
        if right and s.endswith(word, start_pos, end_pos):
            if end_pos is None:
                end_pos = 0
            end_pos -= step
            found = True
        if not found:
            break

    return s[start_pos:end_pos]

def lwstrip(s, word):
    """Strip word only from the left side.
    """
    return wstrip(s, word, right=False)

def rwstrip(s, word):
    """Strip word only from the right side.
    """
    return wstrip(s, word, left=False)


# -*- coding: utf-8 -*-
"""
    pygments.lexers.other
    ~~~~~~~~~~~~~~~~~~~~~

    Just export lexer classes previously contained in this module.

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from testflows._core.contrib.pygments.lexers.sql import SqlLexer, MySqlLexer, SqliteConsoleLexer
from testflows._core.contrib.pygments.lexers.shell import BashLexer, BashSessionLexer, BatchLexer, \
    TcshLexer
from testflows._core.contrib.pygments.lexers.robotframework import RobotFrameworkLexer
from testflows._core.contrib.pygments.lexers.testing import GherkinLexer
from testflows._core.contrib.pygments.lexers.esoteric import BrainfuckLexer, BefungeLexer, RedcodeLexer
from testflows._core.contrib.pygments.lexers.prolog import LogtalkLexer
from testflows._core.contrib.pygments.lexers.snobol import SnobolLexer
from testflows._core.contrib.pygments.lexers.rebol import RebolLexer
from testflows._core.contrib.pygments.lexers.configs import KconfigLexer, Cfengine3Lexer
from testflows._core.contrib.pygments.lexers.modeling import ModelicaLexer
from testflows._core.contrib.pygments.lexers.scripting import AppleScriptLexer, MOOCodeLexer, \
    HybrisLexer
from testflows._core.contrib.pygments.lexers.graphics import PostScriptLexer, GnuplotLexer, \
    AsymptoteLexer, PovrayLexer
from testflows._core.contrib.pygments.lexers.business import ABAPLexer, OpenEdgeLexer, \
    GoodDataCLLexer, MaqlLexer
from testflows._core.contrib.pygments.lexers.automation import AutoItLexer, AutohotkeyLexer
from testflows._core.contrib.pygments.lexers.dsls import ProtoBufLexer, BroLexer, PuppetLexer, \
    MscgenLexer, VGLLexer
from testflows._core.contrib.pygments.lexers.basic import CbmBasicV2Lexer
from testflows._core.contrib.pygments.lexers.pawn import SourcePawnLexer, PawnLexer
from testflows._core.contrib.pygments.lexers.ecl import ECLLexer
from testflows._core.contrib.pygments.lexers.urbi import UrbiscriptLexer
from testflows._core.contrib.pygments.lexers.smalltalk import SmalltalkLexer, NewspeakLexer
from testflows._core.contrib.pygments.lexers.installers import NSISLexer, RPMSpecLexer
from testflows._core.contrib.pygments.lexers.textedit import AwkLexer
from testflows._core.contrib.pygments.lexers.smv import NuSMVLexer

__all__ = []



# Copyright 2020 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import time
import tempfile

parser = re.compile(r".*testflows\.(?P<ppid>\d+)\.(?P<ts>\d+)\.(?P<tss>\d+)\.(?P<pid>\d+)\.log")
glob = "testflows.*.log"


def ppid_glob(ppid):
    return f"testflows.{ppid}.*.log"

def dirname():
    """Return temporary log file directory name.
    """
    return tempfile.gettempdir()

def filename():
    """Return temporary log file name.
    """
    return os.path.join(tempfile.gettempdir(), f"testflows.{os.getppid()}.{time.time():017.7f}.{os.getpid()}.log")



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid
import hashlib

#: debug mode
debug = False
#: time resolution (decimal places)
time_resolution = 6
#: hash length in bytes
hash_length = 8
#: hash function
hash_func = hashlib.sha1
#: disable cli colors
no_colors = False
#: test id
test_id = str(uuid.uuid1())
#: output handler format
output_format = None
#: log file
write_logfile = None
read_logfile = None
#: database
database = None
#: show skipped tests
show_skipped = False
#: show retries
show_retries = False
#: show trimmed result messages
trim = False
#: randomize order of loaded tests
random_order = False
#: global thread pool
global_thread_pool = None
#: global async pool
global_async_pool = None
#: global process pool
global_process_pool = None
#: service timeout
service_timeout = 0.1



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
def transform(file, offset=False, stop=None, bufsize=10485760, encoding=None):
    """Raw read from a file-like object.

    :param file: open file handle
    :param offset: include offset with the message, default: False
    :param stop: stop event
    """
    yield None
    pos = 0

    while True:
        data = file.read(bufsize)
        if encoding and type(data) is bytes:
            data = data.decode(encoding)
        if data:
            if offset:
                yield (data, pos)
                if encoding:
                    pos += len(data.encode(encoding))
                else:
                    pos += len(data)
            else:
                yield data
        if data == "" if encoding else data == b"":
            if stop:
                stop.set()
            yield None
            break



#  Copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Functions for PKCS#1 version 1.5 encryption and signing

This module implements certain functionality from PKCS#1 version 1.5. For a
very clear example, read http://www.di-mgt.com.au/rsa_alg.html#pkcs1schemes

At least 8 bytes of random padding is used when encrypting a message. This makes
these methods much more secure than the ones in the ``rsa`` module.

WARNING: this module leaks information when decryption fails. The exceptions
that are raised contain the Python traceback information, which can be used to
deduce where in the process the failure occurred. DO NOT PASS SUCH INFORMATION
to your users.
"""

import hashlib
import os
import sys
import typing

from . import common, transform, core, key

# ASN.1 codes that describe the hash algorithm used.
HASH_ASN1 = {
    'MD5': b'\x30\x20\x30\x0c\x06\x08\x2a\x86\x48\x86\xf7\x0d\x02\x05\x05\x00\x04\x10',
    'SHA-1': b'\x30\x21\x30\x09\x06\x05\x2b\x0e\x03\x02\x1a\x05\x00\x04\x14',
    'SHA-224': b'\x30\x2d\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x04\x05\x00\x04\x1c',
    'SHA-256': b'\x30\x31\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x01\x05\x00\x04\x20',
    'SHA-384': b'\x30\x41\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x02\x05\x00\x04\x30',
    'SHA-512': b'\x30\x51\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x03\x05\x00\x04\x40',
}

HASH_METHODS = {
    'MD5': hashlib.md5,
    'SHA-1': hashlib.sha1,
    'SHA-224': hashlib.sha224,
    'SHA-256': hashlib.sha256,
    'SHA-384': hashlib.sha384,
    'SHA-512': hashlib.sha512,
}


if sys.version_info >= (3, 6):
    # Python 3.6 introduced SHA3 support.
    HASH_ASN1.update({
        'SHA3-256': b'\x30\x31\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x08\x05\x00\x04\x20',
        'SHA3-384': b'\x30\x41\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x09\x05\x00\x04\x30',
        'SHA3-512': b'\x30\x51\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x0a\x05\x00\x04\x40',
    })

    HASH_METHODS.update({
        'SHA3-256': hashlib.sha3_256,
        'SHA3-384': hashlib.sha3_384,
        'SHA3-512': hashlib.sha3_512,
    })


class CryptoError(Exception):
    """Base class for all exceptions in this module."""


class DecryptionError(CryptoError):
    """Raised when decryption fails."""


class VerificationError(CryptoError):
    """Raised when verification fails."""


def _pad_for_encryption(message: bytes, target_length: int) -> bytes:
    r"""Pads the message for encryption, returning the padded message.

    :return: 00 02 RANDOM_DATA 00 MESSAGE

    >>> block = _pad_for_encryption(b'hello', 16)
    >>> len(block)
    16
    >>> block[0:2]
    b'\x00\x02'
    >>> block[-6:]
    b'\x00hello'

    """

    max_msglength = target_length - 11
    msglength = len(message)

    if msglength > max_msglength:
        raise OverflowError('%i bytes needed for message, but there is only'
                            ' space for %i' % (msglength, max_msglength))

    # Get random padding
    padding = b''
    padding_length = target_length - msglength - 3

    # We remove 0-bytes, so we'll end up with less padding than we've asked for,
    # so keep adding data until we're at the correct length.
    while len(padding) < padding_length:
        needed_bytes = padding_length - len(padding)

        # Always read at least 8 bytes more than we need, and trim off the rest
        # after removing the 0-bytes. This increases the chance of getting
        # enough bytes, especially when needed_bytes is small
        new_padding = os.urandom(needed_bytes + 5)
        new_padding = new_padding.replace(b'\x00', b'')
        padding = padding + new_padding[:needed_bytes]

    assert len(padding) == padding_length

    return b''.join([b'\x00\x02',
                     padding,
                     b'\x00',
                     message])


def _pad_for_signing(message: bytes, target_length: int) -> bytes:
    r"""Pads the message for signing, returning the padded message.

    The padding is always a repetition of FF bytes.

    :return: 00 01 PADDING 00 MESSAGE

    >>> block = _pad_for_signing(b'hello', 16)
    >>> len(block)
    16
    >>> block[0:2]
    b'\x00\x01'
    >>> block[-6:]
    b'\x00hello'
    >>> block[2:-6]
    b'\xff\xff\xff\xff\xff\xff\xff\xff'

    """

    max_msglength = target_length - 11
    msglength = len(message)

    if msglength > max_msglength:
        raise OverflowError('%i bytes needed for message, but there is only'
                            ' space for %i' % (msglength, max_msglength))

    padding_length = target_length - msglength - 3

    return b''.join([b'\x00\x01',
                     padding_length * b'\xff',
                     b'\x00',
                     message])


def encrypt(message: bytes, pub_key: key.PublicKey) -> bytes:
    """Encrypts the given message using PKCS#1 v1.5

    :param message: the message to encrypt. Must be a byte string no longer than
        ``k-11`` bytes, where ``k`` is the number of bytes needed to encode
        the ``n`` component of the public key.
    :param pub_key: the :py:class:`rsa.PublicKey` to encrypt with.
    :raise OverflowError: when the message is too large to fit in the padded
        block.

    >>> from testflows._core.contrib.rsa import key, common
    >>> (pub_key, priv_key) = key.newkeys(256)
    >>> message = b'hello'
    >>> crypto = encrypt(message, pub_key)

    The crypto text should be just as long as the public key 'n' component:

    >>> len(crypto) == common.byte_size(pub_key.n)
    True

    """

    keylength = common.byte_size(pub_key.n)
    padded = _pad_for_encryption(message, keylength)

    payload = transform.bytes2int(padded)
    encrypted = core.encrypt_int(payload, pub_key.e, pub_key.n)
    block = transform.int2bytes(encrypted, keylength)

    return block


def decrypt(crypto: bytes, priv_key: key.PrivateKey) -> bytes:
    r"""Decrypts the given message using PKCS#1 v1.5

    The decryption is considered 'failed' when the resulting cleartext doesn't
    start with the bytes 00 02, or when the 00 byte between the padding and
    the message cannot be found.

    :param crypto: the crypto text as returned by :py:func:`rsa.encrypt`
    :param priv_key: the :py:class:`rsa.PrivateKey` to decrypt with.
    :raise DecryptionError: when the decryption fails. No details are given as
        to why the code thinks the decryption fails, as this would leak
        information about the private key.


    >>> import rsa
    >>> (pub_key, priv_key) = rsa.newkeys(256)

    It works with strings:

    >>> crypto = encrypt(b'hello', pub_key)
    >>> decrypt(crypto, priv_key)
    b'hello'

    And with binary data:

    >>> crypto = encrypt(b'\x00\x00\x00\x00\x01', pub_key)
    >>> decrypt(crypto, priv_key)
    b'\x00\x00\x00\x00\x01'

    Altering the encrypted information will *likely* cause a
    :py:class:`rsa.pkcs1.DecryptionError`. If you want to be *sure*, use
    :py:func:`rsa.sign`.


    .. warning::

        Never display the stack trace of a
        :py:class:`rsa.pkcs1.DecryptionError` exception. It shows where in the
        code the exception occurred, and thus leaks information about the key.
        It's only a tiny bit of information, but every bit makes cracking the
        keys easier.

    >>> crypto = encrypt(b'hello', pub_key)
    >>> crypto = crypto[0:5] + b'X' + crypto[6:] # change a byte
    >>> decrypt(crypto, priv_key)
    Traceback (most recent call last):
    ...
    rsa.pkcs1.DecryptionError: Decryption failed

    """

    blocksize = common.byte_size(priv_key.n)
    encrypted = transform.bytes2int(crypto)
    decrypted = priv_key.blinded_decrypt(encrypted)
    cleartext = transform.int2bytes(decrypted, blocksize)

    # Detect leading zeroes in the crypto. These are not reflected in the
    # encrypted value (as leading zeroes do not influence the value of an
    # integer). This fixes CVE-2020-13757.
    if len(crypto) > blocksize:
        raise DecryptionError('Decryption failed')

    # If we can't find the cleartext marker, decryption failed.
    if cleartext[0:2] != b'\x00\x02':
        raise DecryptionError('Decryption failed')

    # Find the 00 separator between the padding and the message
    try:
        sep_idx = cleartext.index(b'\x00', 2)
    except ValueError:
        raise DecryptionError('Decryption failed')

    return cleartext[sep_idx + 1:]


def sign_hash(hash_value: bytes, priv_key: key.PrivateKey, hash_method: str) -> bytes:
    """Signs a precomputed hash with the private key.

    Hashes the message, then signs the hash with the given key. This is known
    as a "detached signature", because the message itself isn't altered.

    :param hash_value: A precomputed hash to sign (ignores message).
    :param priv_key: the :py:class:`rsa.PrivateKey` to sign with
    :param hash_method: the hash method used on the message. Use 'MD5', 'SHA-1',
        'SHA-224', SHA-256', 'SHA-384' or 'SHA-512'.
    :return: a message signature block.
    :raise OverflowError: if the private key is too small to contain the
        requested hash.

    """

    # Get the ASN1 code for this hash method
    if hash_method not in HASH_ASN1:
        raise ValueError('Invalid hash method: %s' % hash_method)
    asn1code = HASH_ASN1[hash_method]

    # Encrypt the hash with the private key
    cleartext = asn1code + hash_value
    keylength = common.byte_size(priv_key.n)
    padded = _pad_for_signing(cleartext, keylength)

    payload = transform.bytes2int(padded)
    encrypted = priv_key.blinded_encrypt(payload)
    block = transform.int2bytes(encrypted, keylength)

    return block


def sign(message: bytes, priv_key: key.PrivateKey, hash_method: str) -> bytes:
    """Signs the message with the private key.

    Hashes the message, then signs the hash with the given key. This is known
    as a "detached signature", because the message itself isn't altered.

    :param message: the message to sign. Can be an 8-bit string or a file-like
        object. If ``message`` has a ``read()`` method, it is assumed to be a
        file-like object.
    :param priv_key: the :py:class:`rsa.PrivateKey` to sign with
    :param hash_method: the hash method used on the message. Use 'MD5', 'SHA-1',
        'SHA-224', SHA-256', 'SHA-384' or 'SHA-512'.
    :return: a message signature block.
    :raise OverflowError: if the private key is too small to contain the
        requested hash.

    """

    msg_hash = compute_hash(message, hash_method)
    return sign_hash(msg_hash, priv_key, hash_method)


def verify(message: bytes, signature: bytes, pub_key: key.PublicKey) -> str:
    """Verifies that the signature matches the message.

    The hash method is detected automatically from the signature.

    :param message: the signed message. Can be an 8-bit string or a file-like
        object. If ``message`` has a ``read()`` method, it is assumed to be a
        file-like object.
    :param signature: the signature block, as created with :py:func:`rsa.sign`.
    :param pub_key: the :py:class:`rsa.PublicKey` of the person signing the message.
    :raise VerificationError: when the signature doesn't match the message.
    :returns: the name of the used hash.

    """

    keylength = common.byte_size(pub_key.n)
    encrypted = transform.bytes2int(signature)
    decrypted = core.decrypt_int(encrypted, pub_key.e, pub_key.n)
    clearsig = transform.int2bytes(decrypted, keylength)

    # Get the hash method
    method_name = _find_method_hash(clearsig)
    message_hash = compute_hash(message, method_name)

    # Reconstruct the expected padded hash
    cleartext = HASH_ASN1[method_name] + message_hash
    expected = _pad_for_signing(cleartext, keylength)

    if len(signature) != keylength:
        raise VerificationError('Verification failed')

    # Compare with the signed one
    if expected != clearsig:
        raise VerificationError('Verification failed')

    return method_name


def find_signature_hash(signature: bytes, pub_key: key.PublicKey) -> str:
    """Returns the hash name detected from the signature.

    If you also want to verify the message, use :py:func:`rsa.verify()` instead.
    It also returns the name of the used hash.

    :param signature: the signature block, as created with :py:func:`rsa.sign`.
    :param pub_key: the :py:class:`rsa.PublicKey` of the person signing the message.
    :returns: the name of the used hash.
    """

    keylength = common.byte_size(pub_key.n)
    encrypted = transform.bytes2int(signature)
    decrypted = core.decrypt_int(encrypted, pub_key.e, pub_key.n)
    clearsig = transform.int2bytes(decrypted, keylength)

    return _find_method_hash(clearsig)


def yield_fixedblocks(infile: typing.BinaryIO, blocksize: int) -> typing.Iterator[bytes]:
    """Generator, yields each block of ``blocksize`` bytes in the input file.

    :param infile: file to read and separate in blocks.
    :param blocksize: block size in bytes.
    :returns: a generator that yields the contents of each block
    """

    while True:
        block = infile.read(blocksize)

        read_bytes = len(block)
        if read_bytes == 0:
            break

        yield block

        if read_bytes < blocksize:
            break


def compute_hash(message: typing.Union[bytes, typing.BinaryIO], method_name: str) -> bytes:
    """Returns the message digest.

    :param message: the signed message. Can be an 8-bit string or a file-like
        object. If ``message`` has a ``read()`` method, it is assumed to be a
        file-like object.
    :param method_name: the hash method, must be a key of
        :py:const:`HASH_METHODS`.

    """

    if method_name not in HASH_METHODS:
        raise ValueError('Invalid hash method: %s' % method_name)

    method = HASH_METHODS[method_name]
    hasher = method()

    if isinstance(message, bytes):
        hasher.update(message)
    else:
        assert hasattr(message, 'read') and hasattr(message.read, '__call__')
        # read as 1K blocks
        for block in yield_fixedblocks(message, 1024):
            hasher.update(block)

    return hasher.digest()


def _find_method_hash(clearsig: bytes) -> str:
    """Finds the hash method.

    :param clearsig: full padded ASN1 and hash.
    :return: the used hash method.
    :raise VerificationFailed: when the hash method cannot be found
    """

    for (hashname, asn1code) in HASH_ASN1.items():
        if asn1code in clearsig:
            return hashname

    raise VerificationError('Verification failed')


__all__ = ['encrypt', 'decrypt', 'sign', 'verify',
           'DecryptionError', 'VerificationError', 'CryptoError']

if __name__ == '__main__':
    print('Running doctests 1000x or until failure')
    import doctest

    for count in range(1000):
        (failures, tests) = doctest.testmod()
        if failures:
            break

        if count % 100 == 0 and count:
            print('%i times' % count)

    print('Doctests done')



#
# This file is part of pyasn1 software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/pyasn1/license.html
#
import sys

from testflows._core.contrib.pyasn1 import error
from testflows._core.contrib.pyasn1.type import tag
from testflows._core.contrib.pyasn1.type import univ

__all__ = ['NumericString', 'PrintableString', 'TeletexString', 'T61String', 'VideotexString',
           'IA5String', 'GraphicString', 'VisibleString', 'ISO646String',
           'GeneralString', 'UniversalString', 'BMPString', 'UTF8String']

NoValue = univ.NoValue
noValue = univ.noValue


class AbstractCharacterString(univ.OctetString):
    """Creates |ASN.1| schema or value object.

    |ASN.1| class is based on :class:`~pyasn1.type.base.SimpleAsn1Type`,
    its objects are immutable and duck-type Python 2 :class:`str` or Python 3
    :class:`bytes`. When used in octet-stream context, |ASN.1| type assumes
    "|encoding|" encoding.

    Keyword Args
    ------------
    value: :class:`unicode`, :class:`str`, :class:`bytes` or |ASN.1| object
        :class:`unicode` object (Python 2) or :class:`str` (Python 3),
        alternatively :class:`str` (Python 2) or :class:`bytes` (Python 3)
        representing octet-stream of serialised unicode string
        (note `encoding` parameter) or |ASN.1| class instance.
        If `value` is not given, schema object will be created.

    tagSet: :py:class:`~pyasn1.type.tag.TagSet`
        Object representing non-default ASN.1 tag(s)

    subtypeSpec: :py:class:`~pyasn1.type.constraint.ConstraintsIntersection`
        Object representing non-default ASN.1 subtype constraint(s). Constraints
        verification for |ASN.1| type occurs automatically on object
        instantiation.

    encoding: :py:class:`str`
        Unicode codec ID to encode/decode :class:`unicode` (Python 2) or
        :class:`str` (Python 3) the payload when |ASN.1| object is used
        in octet-stream context.

    Raises
    ------
    ~pyasn1.error.ValueConstraintError, ~pyasn1.error.PyAsn1Error
        On constraint violation or bad initializer.
    """

    if sys.version_info[0] <= 2:
        def __str__(self):
            try:
                # `str` is Py2 text representation
                return self._value.encode(self.encoding)

            except UnicodeEncodeError:
                exc = sys.exc_info()[1]
                raise error.PyAsn1UnicodeEncodeError(
                    "Can't encode string '%s' with codec "
                    "%s" % (self._value, self.encoding), exc
                )

        def __unicode__(self):
            return unicode(self._value)

        def prettyIn(self, value):
            try:
                if isinstance(value, unicode):
                    return value
                elif isinstance(value, str):
                    return value.decode(self.encoding)
                elif isinstance(value, (tuple, list)):
                    return self.prettyIn(''.join([chr(x) for x in value]))
                elif isinstance(value, univ.OctetString):
                    return value.asOctets().decode(self.encoding)
                else:
                    return unicode(value)

            except (UnicodeDecodeError, LookupError):
                exc = sys.exc_info()[1]
                raise error.PyAsn1UnicodeDecodeError(
                    "Can't decode string '%s' with codec "
                    "%s" % (value, self.encoding), exc
                )

        def asOctets(self, padding=True):
            return str(self)

        def asNumbers(self, padding=True):
            return tuple([ord(x) for x in str(self)])

    else:
        def __str__(self):
            # `unicode` is Py3 text representation
            return str(self._value)

        def __bytes__(self):
            try:
                return self._value.encode(self.encoding)
            except UnicodeEncodeError:
                exc = sys.exc_info()[1]
                raise error.PyAsn1UnicodeEncodeError(
                    "Can't encode string '%s' with codec "
                    "%s" % (self._value, self.encoding), exc
                )

        def prettyIn(self, value):
            try:
                if isinstance(value, str):
                    return value
                elif isinstance(value, bytes):
                    return value.decode(self.encoding)
                elif isinstance(value, (tuple, list)):
                    return self.prettyIn(bytes(value))
                elif isinstance(value, univ.OctetString):
                    return value.asOctets().decode(self.encoding)
                else:
                    return str(value)

            except (UnicodeDecodeError, LookupError):
                exc = sys.exc_info()[1]
                raise error.PyAsn1UnicodeDecodeError(
                    "Can't decode string '%s' with codec "
                    "%s" % (value, self.encoding), exc
                )

        def asOctets(self, padding=True):
            return bytes(self)

        def asNumbers(self, padding=True):
            return tuple(bytes(self))

    #
    # See OctetString.prettyPrint() for the explanation
    #

    def prettyOut(self, value):
        return value

    def prettyPrint(self, scope=0):
        # first see if subclass has its own .prettyOut()
        value = self.prettyOut(self._value)

        if value is not self._value:
            return value

        return AbstractCharacterString.__str__(self)

    def __reversed__(self):
        return reversed(self._value)


class NumericString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 18)
    )
    encoding = 'us-ascii'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class PrintableString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 19)
    )
    encoding = 'us-ascii'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class TeletexString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 20)
    )
    encoding = 'iso-8859-1'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class T61String(TeletexString):
    __doc__ = TeletexString.__doc__

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class VideotexString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 21)
    )
    encoding = 'iso-8859-1'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class IA5String(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 22)
    )
    encoding = 'us-ascii'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class GraphicString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 25)
    )
    encoding = 'iso-8859-1'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class VisibleString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 26)
    )
    encoding = 'us-ascii'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class ISO646String(VisibleString):
    __doc__ = VisibleString.__doc__

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()

class GeneralString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 27)
    )
    encoding = 'iso-8859-1'

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class UniversalString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 28)
    )
    encoding = "utf-32-be"

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class BMPString(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 30)
    )
    encoding = "utf-16-be"

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()


class UTF8String(AbstractCharacterString):
    __doc__ = AbstractCharacterString.__doc__

    #: Set (on class, not on instance) or return a
    #: :py:class:`~pyasn1.type.tag.TagSet` object representing ASN.1 tag(s)
    #: associated with |ASN.1| type.
    tagSet = AbstractCharacterString.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 12)
    )
    encoding = "utf-8"

    # Optimization for faster codec lookup
    typeId = AbstractCharacterString.getTypeId()



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import testflows._core.cli.arg.type as argtype

from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows._core.transform.log.pipeline import ShortLogPipeline

class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser("short", help="short transform", epilog=epilog(),
            description="Transform log into a short format.",
            formatter_class=HelpFormatter)

        parser.add_argument("input", metavar="input", type=argtype.logfile("r", bufsize=1, encoding="utf-8"),
                nargs="?", help="input log, default: stdin", default="-")
        parser.add_argument("output", metavar="output", type=argtype.file("w", bufsize=1, encoding="utf-8"),
                nargs="?", help='output file, default: stdout', default="-")

        parser.set_defaults(func=cls())

    def handle(self, args):
        ShortLogPipeline(args.input, args.output).run()


# Copyright 2020 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import testflows.settings as settings

from testflows._core.flags import Flags, SKIP
from testflows._core.message import Message

def format_metric(msg, metrics):
    metrics.append(msg)

formatters = {
    Message.METRIC.name: (format_metric,),
}

def transform(metrics):
    """Transform parsed log into metrics.
    """
    line = None
    while True:
        if line is not None:
            msg = line
            formatter = formatters.get(line["message_keyword"], None)
            if formatter:
                flags = Flags(line["test_flags"])
                if flags & SKIP and settings.show_skipped is False:
                    line = None
                else:
                    line = formatter[0](line, *formatter[1:], metrics)
            else:
                line = None
        line = yield line


# Copyright 2020 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import datetime
import testflows._core.cli.arg.type as argtype

from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase

class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser("requirements", help="software requirements specification", epilog=epilog(),
            description="Create new software requirements specification document.",
            formatter_class=HelpFormatter)

        parser.add_argument("output", metavar="output", type=argtype.file("w", bufsize=1, encoding="utf-8"),
            nargs="?", help='output file, default: stdout', default="-")
        parser.add_argument("--number", metavar="number", type=str, help="document number", default="001")
        parser.add_argument("--title", metavar="name", type=str, help="document title", default="Template")
        parser.add_argument("--author", metavar="name", type=str, help="name of the author", default="*[author]*")
        parser.add_argument("--date", metavar="date", type=str, help="document date", default=f"{datetime.datetime.now():%B %d, %Y}")
        parser.set_defaults(func=cls())

    def handle(self, args):
        template_dir = os.path.join(os.path.dirname(__file__), "..", "..", "..", "..", "..", "document", "new")

        with open(os.path.join(template_dir, "requirements.md")) as fd:
            template = fd.read()

        args.output.write(template.format(
            number=args.number,
            title=args.title,
            author=args.author,
            date=args.date))


# Copyright 2020 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import os
import sys
import time
import lzma
import builtins
import _compression

from lzma import compress, decompress

Compressor = lzma.LZMACompressor
Decompressor = lzma.LZMADecompressor

class TailingDecompressReader(_compression.DecompressReader):

    def __init__(self, *args, **kwargs):
        self._tail = kwargs.pop("tail", True)
        self._tail_sleep = float(kwargs.pop("tail_sleep", 0.15))
        # default compressed file marker '.7zXZ'
        self._COMPRESSED_FILE_MARKER = b"\xfd\x37\x7a\x58\x5a"
        # default uncompressed file marker is start of the message
        self._UNCOMPRESSED_FILE_MARKER = "{\"message_keyword\"".encode("utf-8")

        super(TailingDecompressReader, self).__init__(*args, **kwargs)

    def read(self, size=-1, _tail_sleep=0.15):
        if size < 0:
            return self.readall()

        if not size or self._eof:
            return b""
        data = None  # Default if EOF is encountered
        # Depending on the input data, our call to the decompressor may not
        # return any data. In this case, try again after reading another block.
        while True:
            if self._decompressor.eof:
                self.rawblock = (self._decompressor.unused_data or
                            self._fp.read1(_compression.BUFFER_SIZE))
                if not self.rawblock:
                    if not self._tail:
                        break
                    time.sleep(self._tail_sleep)
                    continue
                # Continue to next stream.
                self._decompressor = self._decomp_factory(
                    **self._decomp_args)
                try:
                    data = self._decompressor.decompress(self.rawblock, size)
                except self._trailing_error:
                    # Trailing data isn't a valid compressed stream; ignore it.
                    break
            else:
                if self._decompressor.needs_input:
                    self.rawblock = self._fp.read1(_compression.BUFFER_SIZE)
                    if not self.rawblock:
                        if not self._tail:
                            break
                        time.sleep(self._tail_sleep)
                        continue
                else:
                    self.rawblock = b""
                try:
                    data = self._decompressor.decompress(self.rawblock, size)
                except lzma.LZMAError as e:
                    # try to find valid compressed block by looking for the compressed file marker
                    # or determine that file is not compressed by finding uncompressed file marker
                    if "Input format not supported by decoder" in str(e):
                        while True:
                            raw_data = self._fp.read1(65536)
                            # abort on EOF if not tailing
                            if not raw_data:
                                if not self._tail:
                                    raise
                            self.rawblock += raw_data
                            # try to find compressed file marker
                            compressed_file_marker_idx = self.rawblock.find(self._COMPRESSED_FILE_MARKER)
                            if compressed_file_marker_idx >= 0:
                                self.rawblock = self.rawblock[compressed_file_marker_idx:]
                                break
                            # try to find uncompressed file marker
                            uncompressed_file_marker_idx = self.rawblock.find(self._UNCOMPRESSED_FILE_MARKER)
                            if uncompressed_file_marker_idx >= 0:
                                self.rawblock = self.rawblock[uncompressed_file_marker_idx:]
                                raise
                        # decompress compressed raw block that we found
                        self._decompressor = self._decomp_factory(**self._decomp_args)
                        return self._decompressor.decompress(self.rawblock, size)
            if data:
                break

        if not data:
            self._eof = True
            self._size = self._pos
            return b""
        self._pos += len(data)
        return data


class CompressedFile(lzma.LZMAFile):
    def __init__(self, filename=None, mode="r", *,
            format=None, check=-1, preset=None, filters=None, tail=False):
        self._fp = None
        self._closefp = False
        self._mode = lzma._MODE_CLOSED
        self._raw_mode = False
        self._tail = tail

        if mode in ("r", "rb"):
            if check != -1:
                raise ValueError("Cannot specify an integrity check "
                                 "when opening a file for reading")
            if preset is not None:
                raise ValueError("Cannot specify a preset compression "
                                 "level when opening a file for reading")
            if format is None:
                format = lzma.FORMAT_AUTO
            mode_code = lzma._MODE_READ
        elif mode in ("w", "wb", "a", "ab", "x", "xb"):
            if format is None:
                format = lzma.FORMAT_XZ
            mode_code = lzma._MODE_WRITE
            self._compressor = lzma.LZMACompressor(format=format, check=check,
                preset=preset, filters=filters)
            self._pos = 0
        else:
            raise ValueError("Invalid mode: {!r}".format(mode))

        if filename is sys.stdin.buffer:
            self._fp = filename
            self._closefp = False
            self._mode = mode_code
        elif isinstance(filename, (str, bytes, os.PathLike)):
            if "b" not in mode:
                mode += "b"
            self._fp = builtins.open(filename, mode)
            self._closefp = True
            self._mode = mode_code
        elif hasattr(filename, "read") or hasattr(filename, "write"):
            self._fp = filename
            self._mode = mode_code
        else:
            raise TypeError("filename must be a str, bytes, file or PathLike object")

        if self._mode == lzma._MODE_READ:
            self.raw = TailingDecompressReader(self._fp, lzma.LZMADecompressor,
                trailing_error=lzma.LZMAError, format=format, filters=filters, tail=self._tail)
            self._buffer = io.BufferedReader(self.raw)

    @property
    def name(self):
        return getattr(self._fp, "name", None)

    def read(self, size=-1):
        self._check_can_read()
        if self._raw_mode:
            return self._fp.read(size)
        try:
            return self._buffer.read(size)
        except lzma.LZMAError as e:
            # fall back to raw mode if we can't decompress data
            if "Input format not supported by decoder" in str(e):
                self._raw_mode = True
                return self.raw.rawblock
            raise

    def read1(self, size=-1):
        self._check_can_read()
        if size < 0:
            size = io.DEFAULT_BUFFER_SIZE
        if self._raw_mode:
            return self._fp.read1(size)
        try:
            return self._buffer.read1(size)
        except lzma.LZMAError as e:
            # fall back to raw mode if we can't decompress data
            if "Input format not supported by decoder" in str(e):
                self._raw_mode = True
                return self.raw.rawblock
            raise


# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from testflows._core.contrib.pygments.lexer import RegexLexer, bygroups
from testflows._core.contrib.pygments.token import *

__all__ = ['TestFlowsLexer']

class TestFlowsLexer(RegexLexer):
    name = 'TestFlows'
    aliases = ['tfs']
    filenames = ['*.tfs']

    tokens = {
        'root': [
            (r'(\s*)(Module|Suite|Feature|Test|Scenario|Example|Outline)( .*)', bygroups(Keyword, Keyword.Namespace, Name.Namespace)),
            (r'(\s*)(Step|Background|Given|When|Then|But|And|By|Finally)', Keyword),
            (r'(\s*)(Arguments|Requirements|Attributes|Tags|Examples)', Name.Label),
            (r'(\s*)OK', Result.OK),
            (r'(\s*)Fail', Result.Fail),
            (r'(\s*)Skip', Result.Skip),
            (r'(\s*)Error', Result.Error),
            (r'(\s*)Null', Result.Null),
            (r'(\s*)(XOK|XFail|XNull|XError)', Result.Xout),
            (r' .*\n', Text),
        ]
    }



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from collections import namedtuple

from testflows._core import __version__
from testflows._core.utils.strip import wstrip
from testflows._core.contrib.arpeggio import RegExMatch as _
from testflows._core.contrib.arpeggio import OneOrMore, ZeroOrMore, EOF, Optional, Not
from testflows._core.contrib.arpeggio import ParserPython as PEGParser
from testflows._core.contrib.arpeggio import PTNodeVisitor, visit_parse_tree
from testflows._core.objects import Specification, Requirement

specification_template = """
%(pyname)s = Specification(
    name=%(name)s, 
    description=%(description)s,
    author=%(author)s,
    date=%(date)s, 
    status=%(status)s, 
    approved_by=%(approved_by)s,
    approved_date=%(approved_date)s,
    approved_version=%(approved_version)s,
    version=%(version)s,
    group=%(group)s,
    type=%(type)s,
    link=%(link)s,
    uid=%(uid)s,
    parent=%(parent)s,
    children=%(children)s,
    headings=%(headings)s,
    requirements=%(requirements)s,
    content=%(content)s)
"""

requirement_template = """
%(pyname)s = Requirement(
    name=%(name)s,
    version=%(version)s,
    priority=%(priority)s,
    group=%(group)s,
    type=%(type)s,
    uid=%(uid)s,
    description=%(description)s,
    link=%(link)s,
    level=%(level)s,
    num=%(num)s)

"""

class Visitor(PTNodeVisitor):
    def __init__(self, source_data, *args, **kwargs):
        self.source_data = source_data
        self.output = (
            "# These requirements were auto generated\n"
            "# from software requirements specification (SRS)\n"
            f"# document by TestFlows v{__version__}.\n"
            "# Do not edit by hand but re-generate instead\n"
            "# using \'tfs requirements generate\' command.\n"
            "from testflows.core import Specification\n"
            "from testflows.core import Requirement\n\n"
            "Heading = Specification.Heading\n\n"
            )
        self.levels = [0]
        self.current_level = 0
        self.requirements = []
        self.headings = []
        self.specification = None
        self.pyname_fmt = re.compile(r"[^a-zA-Z0-9]")
        super(Visitor, self).__init__(*args, **kwargs)

    def visit_line(self, node, children):
        pass

    def process_heading(self, node, children):
        level = node[0].value.count("#")
        # normalize header level
        level -= 1
        if self.current_level < level:
            self.levels = self.levels[:level - 1]
        if len(self.levels) < level:
            self.levels += [0] * (level - len(self.levels))
        self.current_level = level
        self.levels[self.current_level - 1] += 1
        num = '.'.join([str(l) for l in self.levels[:self.current_level]])
        return level, num

    def visit_heading(self, node, children):
        level, num = self.process_heading(node, children)
        name = node.heading_name.value
        self.headings.append(Specification.Heading(name=name, level=level, num=num))

    def visit_specification(self, node, children):
        name = str(node.specification_heading.specification_name.value).strip()
        pyname = re.sub(r"_+", "_", self.pyname_fmt.sub("_", name))
        author = None
        status = None
        version = None
        date = None
        approved_by = None
        approved_date = None
        approved_version = None
        description = None
        type = None
        group = None
        link = None
        uid = None
        parent = None
        children = None

        try:
            date = str(node.specification_date.words).strip()
        except:
            pass
        try:
            author = str(node.specification_author.words).strip()
        except:
            pass
        try:
            version = str(node.specification_version.words).strip()
        except:
            pass
        try:
            status = str(node.specification_approval.specification_approval_status.words).strip()
        except:
            pass
        try:
            approved_version = str(node.specification_approval.specification_approval_version.words).strip()
        except:
            pass
        try:
            approved_date = str(node.specification_approval.specification_approval_date.words).strip()
        except:
            pass
        try:
            approved_by = str(node.specification_approval.specification_approval_by.words).strip()
        except:
            pass

        self.specification = Specification(
            name=name, description=description, author=author,
            date=date, status=status, approved_by=approved_by,
            approved_date=approved_date, approved_version=approved_version,
            version=version, group=group, type=type,
            link=link, uid=uid, parent=parent, children=children, content=None)

    def visit_requirement(self, node, children):
        level, num = self.process_heading(node, children)
        name = node.requirement_heading.requirement_name.value
        version = str(node.requirement_version.word)
        description = None
        group = None
        priority = None
        type = None
        uid = None
        link = None

        try:
            description = "\n".join([f'{"":8}{repr(line.value)}' for lines in node.requirement_description for line in lines])
            description = wstrip(description, f"{'':8}'\\n'\n")
            description = f"(\n{description}\n{'':8})"
        except:
            description = "None"
        try:
            priority = str(node.requirement_priority.word)
        except:
            pass
        try:
            group = str(node.requirement_group.word)
        except:
            pass
        try:
            type = str(node.requirement_type.word)
        except:
            pass
        try:
            uid = str(node.requirement_uid.word)
        except:
            pass

        self.headings.append(Specification.Heading(name=name, level=level, num=num))
        self.requirements.append(Requirement(
            name=name, version=version, description=description,
            priority=priority, group=group, type=type,
            uid=uid, link=link, level=level, num=num))

    def visit_document(self, node, children):
        requirements = []

        for rq in self.requirements:
            pyname = re.sub(r"_+", "_", self.pyname_fmt.sub("_", rq.name))

            self.output += requirement_template.lstrip() % {
                "pyname": pyname,
                "name": repr(rq.name),
                "version": repr(rq.version),
                "description": rq.description,
                "priority": repr(rq.priority),
                "group": repr(rq.group),
                "type": repr(rq.type),
                "uid": repr(rq.uid),
                "link": repr(rq.link),
                "level": repr(rq.level),
                "num": repr(rq.num)
            }

            requirements.append(pyname)

        sep = ",\n" + "    " * 2

        self.output += specification_template.lstrip() % {
            "pyname": re.sub(r"_+", "_", self.pyname_fmt.sub("_", self.specification.name)),
            "name": repr(self.specification.name),
            "description": repr(self.specification.description),
            "author": repr(self.specification.author),
            "date": repr(self.specification.date),
            "status": repr(self.specification.status),
            "approved_by": repr(self.specification.approved_by),
            "approved_date": repr(self.specification.approved_date),
            "approved_version": repr(self.specification.approved_version),
            "version": repr(self.specification.version),
            "group": repr(self.specification.group),
            "type": repr(self.specification.type),
            "link": repr(self.specification.link),
            "uid": repr(self.specification.uid),
            "parent": repr(self.specification.parent),
            "children": repr(self.specification.children),
            "headings": f"(\n{'    ' * 2}{sep.join(f'{heading}' for heading in self.headings)}{sep})",
            "requirements": f"(\n{'    ' * 2}{sep.join(rq for rq in requirements)}{sep})",
            "content": "'''\n%s\n'''" % self.source_data.replace("'''", "\'\'\'").rstrip()
        }

        return self.output.rstrip() + "\n"

def Parser():
    """Returns markdown requirements parser.
    """
    def line():
        return _(r"[^\n]*\n?")

    def empty_line():
        return _(r"[ \t]*\n")

    def heading():
        return _(r"#+[ \t]+"), heading_name, _(r"\n")

    def inline_heading():
        return heading()

    def toc_heading():
        return _(r"#+[ \t]+"), _(r"[Tt][Aa][Bb][Ll][Ee] [Oo][Ff] [Cc][Oo][Nn][Tt][Ee][Nn][Tt][Ss]"), _(r"[ \t]*\n")

    def heading_name():
        return _(r"[^\n]+")

    def not_heading():
        return Not(heading)

    def word():
        return _(r"[^\s]+")

    def words():
        return _(r"[^\n]+")

    def requirement_name():
        return _(r"RQ\.[^\n]+")

    def requirement_description():
        return ZeroOrMore((not_heading, line))

    def requirement_heading():
        return _(r"#+[ \t]+"), requirement_name, _(r"\n")

    def requirement_version():
        return _(r"[ \t]*version:[ \t]*"), word

    def requirement_priority():
        return _(r"[ \t]*priority:[ \t]*"), word

    def requirement_type():
        return _(r"[ \t]*type:[ \t]*"), word

    def requirement_group():
        return _(r"[ \t]*group:[ \t]*"), word

    def requirement_uid():
        return _(r"[ \t]*uid:[ \t]*"), word

    def specification_name():
        return _(r"(QA-)?SRS[^\n]+")

    def specification_heading():
        return _(r"#[ \t]+"), specification_name, _(r"\n")

    def specification_author():
        return _(r"\*?\*?[Aa]uthor:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_date():
        return _(r"\*?\*?[Dd]ate:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_version():
        return _(r"\*?\*?[Vv]ersion:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_other():
        return _(r"\*?\*?[^\*\n]+:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_approval_heading():
        return _(r"#+[ \t]+"), _(r"Approval"), _(r"[ \t]*\n")

    def specification_approval_status():
        return _(r"\*?\*?[Ss]tatus:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_approval_version():
        return _(r"\*?\*?[Vv]ersion:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_approval_by():
        return _(r"\*?\*?[Aa]pproved by:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_approval_date():
        return _(r"\*?\*?[Dd]ate:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_approval_other():
        return _(r"\*?\*?[^\*\n]+:\*?\*?[ \t]*"), words, _(r"\n")

    def specification_approval():
        return specification_approval_heading, ZeroOrMore([
            specification_approval_status,
            specification_approval_version,
            specification_approval_by,
            specification_approval_date,
            specification_approval_other,
            empty_line
        ])

    def specification():
        return specification_heading, ZeroOrMore(inline_heading), ZeroOrMore([
            specification_author,
            specification_date,
            specification_version,
            specification_other,
            specification_approval,
            empty_line,
            _(r"[ \t]*[^\*#\n][^\n]*\n")
        ]), [toc_heading, heading]

    def requirement():
        return requirement_heading, requirement_version, ZeroOrMore([
                requirement_priority, requirement_type, requirement_group, requirement_uid
            ]), Optional(requirement_description)

    def document():
        return Optional(OneOrMore([specification, requirement, heading, line])), EOF

    return PEGParser(document, skipws=False)


def generate(source, destination):
    """Generate requirements from markdown source.

    :param source: source file-like object
    :param destination: destination file-like object
    """
    parser = Parser()
    source_data = source.read()
    tree = parser.parse(source_data)
    destination_data = visit_parse_tree(tree, Visitor(source_data=source_data))
    if destination_data:
        destination.write(destination_data)



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import testflows._core.cli.arg.type as argtype

from testflows._core.cli.colors import color
from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows._core.document.srs import Parser, visit_parse_tree
from testflows._core.document.toc import Visitor as VisitorBase
from testflows._core.transform.log.pipeline import ResultsLogPipeline
from testflows._core.utils.timefuncs import strftimedelta

def color_secondary():
    return functools.partial(color, color="white", attrs=["dim"])

def color_primary():
    return functools.partial(color, color="white")

def result_priority(result):
    if result.startswith("X"):
        return 2
    elif result == "OK":
        return 1
    elif result == "Skip":
        return 3
    # Error, Fail, Null
    return 4

def color_result(result):
    if result.startswith("X"):
        return functools.partial(color, color="blue", attrs=["bold"])
    elif result == "OK":
        return functools.partial(color, color="green", attrs=["bold"])
    elif result == "Skip":
        return functools.partial(color, color="cyan", attrs=["bold"])
    # Error, Fail, Null
    return functools.partial(color, color="red", attrs=["bold"])

def color_counts(result):
    if result == "Satisfied":
        return functools.partial(color, color="green", attrs=["bold"])
    elif result == "Unsatisfied":
        return functools.partial(color, color="grey", attrs=["bold"])
    # Unverified
    return functools.partial(color, color="red", attrs=["bold"])

class Counts(object):
    def __init__(self, name, units, ok, nok, untested):
        self.name = name
        self.units = units
        self.ok = ok
        self.nok = nok
        self.untested = untested

    def __bool__(self):
        return self.units > 0

    def __str__(self):
        s = f"{self.units} {self.name if self.units != 1 else self.name.rstrip('s')} ("
        s = color(s, "white", attrs=["bold"])
        r = []
        if self.ok > 0:
            r.append(color_counts("Satisfied")(f"{self.ok} satisfied {(self.ok / self.units) * 100:.1f}%"))
        if self.nok > 0:
            r.append(color_counts("Unsatisfied")(f"{self.nok} unsatisfied {(self.nok / self.units) * 100:.1f}%"))
        if self.untested > 0:
            r.append(color_counts("Untested")(f"{self.untested} untested {(self.untested / self.units) * 100:.1f}%"))
        s += color(", ", "white", attrs=["bold"]).join(r)
        s += color(")\n", "white", attrs=["bold"])
        return s

class Heading(object):
    def __init__(self, name, level, num):
        self.name = name
        self.level = level
        self.num = num

class Requirement(Heading):
    def __init__(self, name, version, uid, level, num):
        self.version = version
        self.uid = uid
        return super(Requirement, self).__init__(name, level, num)

class Visitor(VisitorBase):
    def __init__(self, *args, **kwargs):
        self.headings = []
        super(Visitor, self).__init__(*args, **kwargs)

    def visit_line(self, node, children):
        pass

    def visit_requirement(self, node, children):
        name = node.requirement_heading.requirement_name.value
        description = None
        uid = None
        version = None
        try:
            uid = f"\"{node.uid.word}\""
        except:
            pass
        try:
            version = f"\"{node.version.word}\""
        except:
            pass
        res = self.process_heading(node, children)
        if res:
            level, num = res
            self.headings.append(Requirement(name, version, uid, level, num))

    def visit_heading(self, node, children):
        res = self.process_heading(node, children)
        if res:
            level, num = res
            name = node.heading_name.value
            self.headings.append(Heading(name, level, num))

    def visit_document(self, node, children):
        return self.headings

class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser("srs-coverage", help="SRS (Software Requirements Specification) coverage report", epilog=epilog(),
            description="Generate SRS (Software Requirements Specification) coverage report.",
            formatter_class=HelpFormatter)

        parser.add_argument("srs", metavar="srs", type=argtype.file("r", bufsize=1, encoding="utf-8"),
                nargs=1, help="source file")
        parser.add_argument("input", metavar="input", type=argtype.logfile("r", bufsize=1, encoding="utf-8"),
                nargs="?", help="input log, default: stdin", default="-")
        parser.add_argument("output", metavar="output", type=argtype.file("w", bufsize=1, encoding="utf-8"),
                nargs="?", help='output file, default: stdout', default="-")
        parser.add_argument("--only", metavar="status", type=str, nargs="+", help="verification status",
            choices=["satisfied", "unsatisfied", "untested"],
            default=["satisfied", "unsatisfied", "untested"])
        parser.set_defaults(func=cls())

    def generate(self, output, headings, tested, only):
        counts = Counts("requirements", *([0] * 4))

        for heading in headings:
            indent = "  " * (heading.level - 1)
            if isinstance(heading, Requirement):
                counts.units += 1
                if tested.get(heading.name) is None:
                    counts.untested += 1
                    if "untested" in only:
                        output.write(color(f"{indent}\u270E ", "grey", attrs=["bold"]) + color_primary()(f"{heading.num} {heading.name}\n"))
                        output.write(color(f"{indent}  no tests\n", "white", attrs=["dim"]))
                else:
                    _tests = []
                    _color = None
                    _priority = 0
                    for test, result in tested.get(heading.name):
                        _tests.append(f"{indent}  [ {color_result(result['result_type'])(result['result_type'])} ] "
                            f"{color_secondary()(strftimedelta(result['message_rtime']))} "
                            f"{color_secondary()(test['test_name'])}\n")
                        if result_priority(result["result_type"]) > _priority:
                            _color = color_result(result["result_type"])
                            _priority = result_priority(result["result_type"])
                    icon = "\u2714"
                    _include = True
                    if _priority > 2:
                        icon = "\u2718"
                        counts.nok += 1
                        if "unsatisfied" not in only:
                            _include = False
                    else:
                        if "satisfied" not in only:
                            _include = False
                        counts.ok += 1

                    if _include:
                        output.write(_color(f"{indent}{icon} ") + color_primary()(f"{heading.num} {heading.name}\n") + "".join(_tests))

            else:
                output.write(color(f"{indent}\u27e5 {heading.num} {heading.name}\n", "white", attrs=["dim"]))

        # print summary
        output.write(f"\n{counts}\n")

    def handle(self, args):
        parser = Parser()
        with args.srs[0] as requirements:
            requirements_data = requirements.read()

        tree = parser.parse(requirements_data)
        headings = visit_parse_tree(tree, Visitor())
        results = {}
        ResultsLogPipeline(args.input, results).run()
        # map result requirements to tests
        tested = {}

        for result in results["tests"]:
            test, result = results["tests"][result].values()
            for req in test["requirements"]:
                if tested.get(req["requirement_name"]) is None:
                    tested[req["requirement_name"]] = []
                tested[req["requirement_name"]].append((test, result))
        # generate report
        with args.output as output:
            self.generate(output, headings, tested, args.only)




# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import testflows.settings as settings

from testflows._core.flags import Flags, SKIP
from testflows._core.cli.colors import color
from testflows._core.message import Message
from testflows._core.testtype import TestType

width = 70
count = 0

def color_result(result):
    if result.startswith("X"):
        return color(".", "blue", attrs=["bold"])
    elif result == "OK":
        return color(".", "green", attrs=["bold"])
    elif result == "Skip":
        return color("-", "white", attrs=["dim"])
    # Error, Fail, Null
    elif result == "Error":
        return color("E", "yellow", attrs=["bold"])
    elif result == "Fail":
         return color("F", "red", attrs=["bold"])
    elif result == "Null":
        return color("N", "magenta", attrs=["bold"])
    else:
        raise ValueError(f"unknown result {result}")

def format_prompt(msg):
    global count
    lines = (msg["message"] or "").splitlines()
    icon = "\u270d  "
    if msg["message"].startswith("Paused"):
        icon = "\u270b "
    out = ""
    if count > 0:
        out += "\n"
    out += color(icon + lines[0], "yellow", attrs=["bold"])
    if len(lines) > 1:
        out += "\n" + color("\n".join(lines[1:]), "white", attrs=["dim"])
    count = 0
    return out

def format_input(msg):
    global count
    out = color(msg['message'], "white") + "\n"
    count = 0
    return out

def format_result(msg):
    global count
    flags = Flags(msg["test_flags"])
    if flags & SKIP and settings.show_skipped is False:
        return

    if getattr(TestType, msg["test_type"]) < TestType.Iteration:
        return

    count += 1
    _result = f"{color_result(msg['result_type'])}"
    # wrap if we hit max width
    if count >= width:
        count = 0
        _result += "\n"

    return _result

formatters = {
    Message.INPUT.name: (format_input,),
    Message.PROMPT.name: (format_prompt,),
    Message.RESULT.name: (format_result,)
}

def transform(stop_event, show_input=True):
    """Transform parsed log line into a short format.
    """
    line = None
    while True:
        if line is not None:
            formatter = formatters.get(line["message_keyword"], None)
            if formatter:
                if formatter[0] is format_input and show_input is False:
                    line = None
                else:
                    line = formatter[0](line, *formatter[1:])
                    n = 0
            else:
                line = None

        if stop_event.is_set():
            if line is None:
                line = ""
            line += "\n"
        line = yield line



#  Copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Functions for generating random numbers."""

# Source inspired by code by Yesudeep Mangalapilly <yesudeep@gmail.com>

import os
import struct

from .common import bit_size as get_bit_size
from .transform import bytes2int


def read_random_bits(nbits: int) -> bytes:
    """Reads 'nbits' random bits.

    If nbits isn't a whole number of bytes, an extra byte will be appended with
    only the lower bits set.
    """

    nbytes, rbits = divmod(nbits, 8)

    # Get the random bytes
    randomdata = os.urandom(nbytes)

    # Add the remaining random bits
    if rbits > 0:
        randomvalue = ord(os.urandom(1))
        randomvalue >>= (8 - rbits)
        randomdata = struct.pack("B", randomvalue) + randomdata

    return randomdata


def read_random_int(nbits: int) -> int:
    """Reads a random integer of approximately nbits bits.
    """

    randomdata = read_random_bits(nbits)
    value = bytes2int(randomdata)

    # Ensure that the number is large enough to just fill out the required
    # number of bits.
    value |= 1 << (nbits - 1)

    return value


def read_random_odd_int(nbits: int) -> int:
    """Reads a random odd integer of approximately nbits bits.

    >>> read_random_odd_int(512) & 1
    1
    """

    value = read_random_int(nbits)

    # Make sure it's odd
    return value | 1


def randint(maxvalue: int) -> int:
    """Returns a random integer x with 1 <= x <= maxvalue

    May take a very long time in specific situations. If maxvalue needs N bits
    to store, the closer maxvalue is to (2 ** N) - 1, the faster this function
    is.
    """

    bit_size = get_bit_size(maxvalue)

    tries = 0
    while True:
        value = read_random_int(bit_size)
        if value <= maxvalue:
            break

        if tries % 10 == 0 and tries:
            # After a lot of tries to get the right number of bits but still
            # smaller than maxvalue, decrease the number of bits by 1. That'll
            # dramatically increase the chances to get a large enough number.
            bit_size -= 1
        tries += 1

    return value



# -*- coding: utf-8 -*-
"""
    pygments.styles.monokai
    ~~~~~~~~~~~~~~~~~~~~~~~

    Mimic the Monokai color scheme. Based on tango.py.

    http://www.monokai.nl/blog/2006/07/15/textmate-color-theme/

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from testflows._core.contrib.pygments.style import Style
from testflows._core.contrib.pygments.token import Keyword, Name, Comment, String, Error, Text, \
     Number, Operator, Generic, Whitespace, Punctuation, Other, Literal

class MonokaiStyle(Style):
    """
    This style mimics the Monokai color scheme.
    """

    background_color = "#272822"
    highlight_color = "#49483e"

    styles = {
        # No corresponding class for the following:
        Text:                      "#f8f8f2", # class:  ''
        Whitespace:                "",        # class: 'w'
        Error:                     "#960050 bg:#1e0010", # class: 'err'
        Other:                     "",        # class 'x'

        Comment:                   "#75715e", # class: 'c'
        Comment.Multiline:         "",        # class: 'cm'
        Comment.Preproc:           "",        # class: 'cp'
        Comment.Single:            "",        # class: 'c1'
        Comment.Special:           "",        # class: 'cs'

        Keyword:                   "#66d9ef", # class: 'k'
        Keyword.Constant:          "",        # class: 'kc'
        Keyword.Declaration:       "",        # class: 'kd'
        Keyword.Namespace:         "#f92672", # class: 'kn'
        Keyword.Pseudo:            "",        # class: 'kp'
        Keyword.Reserved:          "",        # class: 'kr'
        Keyword.Type:              "",        # class: 'kt'

        Operator:                  "#f92672", # class: 'o'
        Operator.Word:             "",        # class: 'ow' - like keywords

        Punctuation:               "#f8f8f2", # class: 'p'

        Name:                      "#f8f8f2", # class: 'n'
        Name.Attribute:            "#a6e22e", # class: 'na' - to be revised
        Name.Builtin:              "",        # class: 'nb'
        Name.Builtin.Pseudo:       "",        # class: 'bp'
        Name.Class:                "#a6e22e", # class: 'nc' - to be revised
        Name.Constant:             "#66d9ef", # class: 'no' - to be revised
        Name.Decorator:            "#a6e22e", # class: 'nd' - to be revised
        Name.Entity:               "",        # class: 'ni'
        Name.Exception:            "#a6e22e", # class: 'ne'
        Name.Function:             "#a6e22e", # class: 'nf'
        Name.Property:             "",        # class: 'py'
        Name.Label:                "",        # class: 'nl'
        Name.Namespace:            "",        # class: 'nn' - to be revised
        Name.Other:                "#a6e22e", # class: 'nx'
        Name.Tag:                  "#f92672", # class: 'nt' - like a keyword
        Name.Variable:             "",        # class: 'nv' - to be revised
        Name.Variable.Class:       "",        # class: 'vc' - to be revised
        Name.Variable.Global:      "",        # class: 'vg' - to be revised
        Name.Variable.Instance:    "",        # class: 'vi' - to be revised

        Number:                    "#ae81ff", # class: 'm'
        Number.Float:              "",        # class: 'mf'
        Number.Hex:                "",        # class: 'mh'
        Number.Integer:            "",        # class: 'mi'
        Number.Integer.Long:       "",        # class: 'il'
        Number.Oct:                "",        # class: 'mo'

        Literal:                   "#ae81ff", # class: 'l'
        Literal.Date:              "#e6db74", # class: 'ld'

        String:                    "#e6db74", # class: 's'
        String.Backtick:           "",        # class: 'sb'
        String.Char:               "",        # class: 'sc'
        String.Doc:                "",        # class: 'sd' - like a comment
        String.Double:             "",        # class: 's2'
        String.Escape:             "#ae81ff", # class: 'se'
        String.Heredoc:            "",        # class: 'sh'
        String.Interpol:           "",        # class: 'si'
        String.Other:              "",        # class: 'sx'
        String.Regex:              "",        # class: 'sr'
        String.Single:             "",        # class: 's1'
        String.Symbol:             "",        # class: 'ss'


        Generic:                   "",        # class: 'g'
        Generic.Deleted:           "#f92672", # class: 'gd',
        Generic.Emph:              "italic",  # class: 'ge'
        Generic.Error:             "",        # class: 'gr'
        Generic.Heading:           "",        # class: 'gh'
        Generic.Inserted:          "#a6e22e", # class: 'gi'
        Generic.Output:            "#66d9ef", # class: 'go'
        Generic.Prompt:            "bold #f92672", # class: 'gp'
        Generic.Strong:            "bold",    # class: 'gs'
        Generic.Subheading:        "#75715e", # class: 'gu'
        Generic.Traceback:         "",        # class: 'gt'
    }
